@@ -71,7 +71,9 @@ cabase <- function(cls,ovf,estf, estcovf=NULL,findmean=TRUE,
    ovout <- ### if (cacall) clusterEvalQ(cls,ovf(z168)) else
                         clusterEvalQ(cls,ovf(z168)) 
    # theta-hats, with the one for chunk i in row i
-   thts <- t(sapply(ovout,estf))
+   # thts <- t(sapply(ovout,estf))
+   tmp <- lapply(ovout, estf)
+   thts <- Reduce(rbind,tmp)
    if (is.vector(thts)) thts <- matrix(thts,ncol=1)
    # res will be the returned result of this function
    res <- list()
@@ -103,6 +105,10 @@ cabase <- function(cls,ovf,estf, estcovf=NULL,findmean=TRUE,
          thtcov <- matrix(0,nrow=lth,ncol=lth)
          for (i in 1:nchunks) {
             summand <- estcovf(ovout[[i]]) 
+            if (any(dim(thtcov) != dim(summand))) {
+               print('dimension mismatch')
+               stop('likely cause is constant variable in some chunk')
+            }
             thtcov <- thtcov + summand
          }
          res$thtcov <- (thtcov / nchunks) / nchunks
@@ -117,11 +123,9 @@ cabase <- function(cls,ovf,estf, estcovf=NULL,findmean=TRUE,
 # ca() wrapper for lm()
 #
 # arguments:
-#
 #    cls: cluster
 #    lmargs: quoted string representing arguments to lm()
 #            e.g. "weight ~ height + age, data-mlb"
-#
 # value: Software Alchemy estimate, statistically equivalent to direct
 #    nonparallel call to lm(); R list is value of ca() 
 #
@@ -144,11 +148,54 @@ caglm <- function(cls,glmargs) {
    cabase(cls,ovf,coef,vcov)
 }
 
+# ca() wrapper for knnest(); 
+# arguments:
+#    cls: cluster
+#    yname: name of distributed Y vector
+#    k: number of nearest neighbors
+#    xname: if nonempty, this is the distributed matrix of X values, 
+#           which will be fed into preprocessx() to produce the 
+#           global variable xdata at the nodes; if empty, xdata will be
+#           generated
+# value:
+#    none at caller; xdata, kout are left there at the nodes 
+#    for future use
+caknn <- function(cls,yname,k,xname='') {
+   if (xname != '') {
+      # run preprocessx() to generate xdata
+      cmd <- paste('xdata <<- preprocessx(',xname,',',k,')',sep='')
+      doclscmd(cls,cmd)
+   }
+   cmd <- paste('kout <<- knnest(',yname,',xdata,',k,')',sep='')
+   doclscmd(cls,cmd)
+}
+
+# needs fixing
+### # kNN predict
+### # arguments:
+### #    predpts: matrix/df of X values at which to find est. reg. ftn.;
+### #             if NULL, it is assumed that predpts already exists at the
+### #             nodes
+### # value:
+### #    est. reg. ftn. value at those points
+### # assumes kout present at the nodes
+### caknnpred <- function(cls,predpts) {
+###    if (!is.matrix(predpts)) 
+###       predpts <- as.matrix(predpts)
+###    clusterEvalQ(cls,library(regtools))
+###    if (!is.null(predpts))
+###       clusterExport(cls,'predpts',envir=environment())
+###    predys <- doclscmd(cls,'predy <<- predict(kout,predpts)')
+###    predys <- Reduce(rbind,predys)
+###    colMeans(predys)
+### }
+
 # ca() wrapper for prcomp()
 # arguments:
 #
 #    prcompargs: arguments to go into prcomp()
-#    p: number of variables
+#    p: number of variables (cannot be inferred by the code, since
+#       prcompargs can be so general
 #
 # value:
 #
